import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from sklearn.decomposition import PCA
from scipy.cluster import hierarchy

def pca_plot(venomseq):
  pca = PCA(n_components=2)
  X_r = pca.fit(venomseq.connectivity.wcs.T).transform(venomseq.connectivity.wcs.T)
  f, ax = plt.subplots(1, 1)
  ax.scatter(X_r[:,0], X_r[:,1], s=12)
  return ax

def heatmap(matrix, annotations=None):
  sns.clustermap(data=matrix)

# def heatmap_wrapper(data, vmin=None, vmax=None, cmap=None, center=None, robust=False,
#                     annot=None, fmt=".2g", annot_kws=None,
#                     linewidths=0, linecolor="white",
#                     cbar=True, cbar_kws=None, cbar_ax=None,
#                     square=False, xticklabels="auto", yticklabels="auto",
#                     mask=None, ax=None, **kwargs):

#   # Initialize the plotter object
#   plotter = _HeatMapper(data, vmin, vmax, cmap, center, robust, annot, fmt,
#                         annot_kws, cbar, cbar_kws, xticklabels,
#                         yticklabels, mask)

#   # Add the pcolormesh kwargs here
#   kwargs["linewidths"] = linewidths
#   kwargs["edgecolor"] = linecolor

#   # Draw the plot and return the Axes
#   if ax is None:
#     ax = plt.gca()
#   if square:
#     ax.set_aspect("equal")
#   plotter.plot(ax, cbar_ax, kwargs)
#   return ax

# class _HeatMapper(object):
#   """Draw a heatmap plot of a matrix with nice labels and colormaps."""

#   def __init__(self, data, vmin, vmax, cmap, center, robust, annot, fmt,
#          annot_kws, cbar, cbar_kws,
#          xticklabels=True, yticklabels=True, mask=None):
#     """Initialize the plotting object."""
#     # We always want to have a DataFrame with semantic information
#     # and an ndarray to pass to matplotlib
#     if isinstance(data, pd.DataFrame):
#       plot_data = data.values
#     else:
#       plot_data = np.asarray(data)
#       data = pd.DataFrame(plot_data)

#     # Validate the mask and convet to DataFrame
#     mask = _matrix_mask(data, mask)

#     plot_data = np.ma.masked_where(np.asarray(mask), plot_data)

#     # Get good names for the rows and columns
#     xtickevery = 1
#     if isinstance(xticklabels, int):
#       xtickevery = xticklabels
#       xticklabels = _index_to_ticklabels(data.columns)
#     elif xticklabels is True:
#       xticklabels = _index_to_ticklabels(data.columns)
#     elif xticklabels is False:
#       xticklabels = []

#     ytickevery = 1
#     if isinstance(yticklabels, int):
#       ytickevery = yticklabels
#       yticklabels = _index_to_ticklabels(data.index)
#     elif yticklabels is True:
#       yticklabels = _index_to_ticklabels(data.index)
#     elif yticklabels is False:
#       yticklabels = []

#     # Get the positions and used label for the ticks
#     nx, ny = data.T.shape

#     if not len(xticklabels):
#       self.xticks = []
#       self.xticklabels = []
#     elif isinstance(xticklabels, string_types) and xticklabels == "auto":
#       self.xticks = "auto"
#       #self.xticklabels = _index_to_ticklabels(data.columns)
#       #self.xticklabels = np.array(data.columns) + 1
#       reordered_species_names = []
#       for x in np.array(data.columns):
#         reordered_species_names.append(species_names[x])
#       self.xticklabels = np.array(reordered_species_names)
#     else:
#       self.xticks, self.xticklabels = self._skip_ticks(xticklabels,
#                                xtickevery)

#     if not len(yticklabels):
#       self.yticks = []
#       self.yticklabels = []
#     elif isinstance(yticklabels, string_types) and yticklabels == "auto":
#       self.yticks = "auto"
#       self.yticklabels = _index_to_ticklabels(data.index)
#     else:
#       self.yticks, self.yticklabels = self._skip_ticks(yticklabels,
#                                ytickevery)

#     # Get good names for the axis labels
#     xlabel = _index_to_label(data.columns)
#     ylabel = _index_to_label(data.index)
#     self.xlabel = xlabel if xlabel is not None else ""
#     self.ylabel = ylabel if ylabel is not None else ""

#     # Determine good default values for the colormapping
#     self._determine_cmap_params(plot_data, vmin, vmax,
#                   cmap, center, robust)

#     # Sort out the annotations
#     if annot is None:
#       annot = False
#       annot_data = None
#     elif isinstance(annot, bool):
#       if annot:
#         annot_data = plot_data
#       else:
#         annot_data = None
#     else:
#       try:
#         annot_data = annot.values
#       except AttributeError:
#         annot_data = annot
#       if annot.shape != plot_data.shape:
#         raise ValueError('Data supplied to "annot" must be the same '
#                  'shape as the data to plot.')
#       annot = True

#     # Save other attributes to the object
#     self.data = data
#     self.plot_data = plot_data

#     self.annot = annot
#     self.annot_data = annot_data

#     self.fmt = fmt
#     self.annot_kws = {} if annot_kws is None else annot_kws
#     self.cbar = cbar
#     self.cbar_kws = {} if cbar_kws is None else cbar_kws
#     self.cbar_kws.setdefault('ticks', mpl.ticker.MaxNLocator(6))

#   def _determine_cmap_params(self, plot_data, vmin, vmax,
#                  cmap, center, robust):
#     """Use some heuristics to set good defaults for colorbar and range."""
#     calc_data = plot_data.data[~np.isnan(plot_data.data)]
#     if vmin is None:
#       vmin = np.percentile(calc_data, 2) if robust else calc_data.min()
#     if vmax is None:
#       vmax = np.percentile(calc_data, 98) if robust else calc_data.max()
#     self.vmin, self.vmax = vmin, vmax

#     # Choose default colormaps if not provided
#     if cmap is None:
#       #if center is None:
#       #    self.cmap = cm.inferno
#       #else:
#       #    self.cmap = cm.icefire
#       self.cmap = cm.coolwarm
#     elif isinstance(cmap, string_types):
#       self.cmap = mpl.cm.get_cmap(cmap)
#     elif isinstance(cmap, list):
#       self.cmap = mpl.colors.ListedColormap(cmap)
#     else:
#       self.cmap = cmap

#     # Recenter a divergent colormap
#     if center is not None:
#       vrange = max(vmax - center, center - vmin)
#       normlize = mpl.colors.Normalize(center - vrange, center + vrange)
#       cmin, cmax = normlize([vmin, vmax])
#       cc = np.linspace(cmin, cmax, 256)
#       self.cmap = mpl.colors.ListedColormap(self.cmap(cc))

#   def _annotate_heatmap(self, ax, mesh):
#     """Add textual labels with the value in each cell."""
#     mesh.update_scalarmappable()
#     height, width = self.annot_data.shape
#     xpos, ypos = np.meshgrid(np.arange(width) + .5, np.arange(height) + .5)
#     for x, y, m, color, val in zip(xpos.flat, ypos.flat,
#                      mesh.get_array(), mesh.get_facecolors(),
#                      self.annot_data.flat):
#       if m is not np.ma.masked:
#         lum = relative_luminance(color)
#         text_color = ".15" if lum > .408 else "w"
#         annotation = ("{:" + self.fmt + "}").format(val)
#         text_kwargs = dict(color=text_color, ha="center", va="center")
#         text_kwargs.update(self.annot_kws)
#         ax.text(x, y, annotation, **text_kwargs)

#   def _skip_ticks(self, labels, tickevery):
#     """Return ticks and labels at evenly spaced intervals."""
#     n = len(labels)
#     if tickevery == 0:
#       ticks, labels = [], []
#     elif tickevery == 1:
#       ticks, labels = np.arange(n) + .5, labels
#     else:
#       start, end, step = 0, n, tickevery
#       ticks = np.arange(start, end, step) + .5
#       labels = labels[start:end:step]
#     return ticks, labels

#   def _auto_ticks(self, ax, labels, axis):
#     """Determine ticks and ticklabels that minimize overlap."""
#     transform = ax.figure.dpi_scale_trans.inverted()
#     bbox = ax.get_window_extent().transformed(transform)
#     size = [bbox.width, bbox.height][axis]
#     axis = [ax.xaxis, ax.yaxis][axis]
#     tick, = axis.set_ticks([0])
#     fontsize = tick.label.get_size()
#     max_ticks = int(size // (fontsize / 72))
#     if max_ticks < 1:
#       return [], []
#     tick_every = len(labels) // max_ticks + 1
#     tick_every = 1 if tick_every == 0 else tick_every
#     ticks, labels = self._skip_ticks(labels, tick_every)
#     return ticks, labels

#   def plot(self, ax, cax, kws):
#     """Draw the heatmap on the provided Axes."""
#     # Remove all the Axes spines
#     despine(ax=ax, left=True, bottom=True)

#     # Draw the heatmap
#     mesh = ax.pcolormesh(self.plot_data, vmin=self.vmin, vmax=self.vmax,
#                cmap=self.cmap, **kws)

#     # Set the axis limits
#     ax.set(xlim=(0, self.data.shape[1]), ylim=(0, self.data.shape[0]))

#     # Invert the y axis to show the plot in matrix form
#     ax.invert_yaxis()

#     # Possibly add a colorbar
#     if self.cbar:
#       cb = ax.figure.colorbar(mesh, cax, ax, **self.cbar_kws)
#       cb.outline.set_linewidth(0)
#       # If rasterized is passed to pcolormesh, also rasterize the
#       # colorbar to avoid white lines on the PDF rendering
#       if kws.get('rasterized', False):
#         cb.solids.set_rasterized(True)

#     # Add row and column labels
#     if isinstance(self.xticks, string_types) and self.xticks == "auto":
#       xticks, xticklabels = self._auto_ticks(ax, self.xticklabels, 0)
#     else:
#       xticks, xticklabels = self.xticks, self.xticklabels

#     if isinstance(self.yticks, string_types) and self.yticks == "auto":
#       yticks, yticklabels = self._auto_ticks(ax, self.yticklabels, 1)
#     else:
#       yticks, yticklabels = self.yticks, self.yticklabels

#     ax.set(xticks=xticks, yticks=yticks)
#     xtl = ax.set_xticklabels(xticklabels)
#     ytl = ax.set_yticklabels(yticklabels, rotation="vertical")

#     # Possibly rotate them if they overlap
#     if hasattr(ax.figure.canvas, "get_renderer"):
#       ax.figure.draw(ax.figure.canvas.get_renderer())
#     if axis_ticklabels_overlap(xtl):
#       plt.setp(xtl, rotation="vertical")
#     if axis_ticklabels_overlap(ytl):
#       plt.setp(ytl, rotation="horizontal")

#     # Add the axis labels
#     ax.set(xlabel=self.xlabel, ylabel=self.ylabel)

#     # Annotate the cells with the formatted values
#     if self.annot:
#       self._annotate_heatmap(ax, mesh)



# def _matrix_mask(data, mask):
#   if mask is None:
#     mask = np.zeros(data.shape, np.bool)
#   if isinstance(mask, np.ndarray):
#     if mask.shape != data.shape:
#       raise ValueError("Mask must have the same shape as data.")
#     mask = pd.DataFrame(mask,
#               index=data.index,
#               columns=data.columns,
#               dtype=np.bool)
#   elif isinstance(mask, pd.DataFrame):
#     if not mask.index.equals(data.index) \
#        and mask.columns.equals(data.columns):
#       err = "Mask must have the same index and columns as data."
#       raise ValueError(err)
#   mask = mask | pd.isnull(data)
#   return mask

# def _index_to_label(index):
#   if isinstance(index, pd.MultiIndex):
#     return "-".join(map(to_utf8, index.names))
#   else:
#     return index.name
# def _index_to_ticklabels(index):
#   if isinstance(index, pd.MultiIndex):
#     return ["-".join(map(to_utf8, i)) for i in index.values]
#   else:
#     return index.values
# string_types = str
# def despine(fig=None, ax=None, top=True, right=True, left=False,
#       bottom=False, offset=None, trim=False):
#   if fig is None and ax is None:
#     axes = plt.gcf().axes
#   elif fig is not None:
#     axes = fig.axes
#   elif ax is not None:
#     axes = [ax]
#   for ax_i in axes:
#     for side in ["top", "right", "left", "bottom"]:
#       is_visible = not locals()[side]
#       ax_i.spines[side].set_visible(is_visible)
#       if offset is not None and is_visible:
#         try:
#           val = offset.get(side, 0)
#         except AttributeError:
#           val = offset
#         _set_spine_position(ax_i.spines[side], ('outward', val))
#     if left and not right:
#       maj_on = any(t.tick1On for t in ax_i.yaxis.majorTicks)
#       min_on = any(t.tick1On for t in ax_i.yaxis.minorTicks)
#       ax_i.yaxis.set_ticks_position("right")
#       for t in ax_i.yaxis.majorTicks:
#         t.tick2On = maj_on
#       for t in ax_i.yaxis.minorTicks:
#         t.tick2On = min_on
#     if bottom and not top:
#       maj_on = any(t.tick1On for t in ax_i.xaxis.majorTicks)
#       min_on = any(t.tick1On for t in ax_i.xaxis.minorTicks)
#       ax_i.xaxis.set_ticks_position("top")
#       for t in ax_i.xaxis.majorTicks:
#         t.tick2On = maj_on
#       for t in ax_i.xaxis.minorTicks:
#         t.tick2On = min_on
#     if trim:
#       xticks = ax_i.get_xticks()
#       if xticks.size:
#         firsttick = np.compress(xticks >= min(ax_i.get_xlim()),
#                     xticks)[0]
#         lasttick = np.compress(xticks <= max(ax_i.get_xlim()),
#                      xticks)[-1]
#         ax_i.spines['bottom'].set_bounds(firsttick, lasttick)
#         ax_i.spines['top'].set_bounds(firsttick, lasttick)
#         newticks = xticks.compress(xticks <= lasttick)
#         newticks = newticks.compress(newticks >= firsttick)
#         ax_i.set_xticks(newticks)
#       yticks = ax_i.get_yticks()
#       if yticks.size:
#         firsttick = np.compress(yticks >= min(ax_i.get_ylim()),
#                     yticks)[0]
#         lasttick = np.compress(yticks <= max(ax_i.get_ylim()),
#                      yticks)[-1]
#         ax_i.spines['left'].set_bounds(firsttick, lasttick)
#         ax_i.spines['right'].set_bounds(firsttick, lasttick)
#         newticks = yticks.compress(yticks <= lasttick)
#         newticks = newticks.compress(newticks >= firsttick)
#         ax_i.set_yticks(newticks)
# def axis_ticklabels_overlap(labels):
#   if not labels:
#     return False
#   try:
#     bboxes = [l.get_window_extent() for l in labels]
#     overlaps = [b.count_overlaps(bboxes) for b in bboxes]
#     return max(overlaps) > 1
#   except RuntimeError:
#     return False


from collections import namedtuple
from glob import glob
import pandas as pd
import numpy as np
import os
import json
from scipy import stats
import itertools
from tqdm import tqdm

from .utils import read_gctx
from .visualizations import *

import ipdb

SYMBOL_MAP_FNAME = os.path.join(os.path.dirname(__file__), "data", "symbol_map.npy")
DEFAULT_PCL_FNAME = os.path.join(os.path.dirname(__file__), "data", "pcls.json")

class VenomSeq(object):
  def __init__(self,
               samples_file = None,
               gctx_file = None,
               signatures_dir = None,
               pcls_file = DEFAULT_PCL_FNAME,
               verbose = True):
    #self.counts_file = counts_file
    self.verbose = verbose
    self.samples_file = samples_file  # Metadata describing venom samples
    self.gctx_file = gctx_file
    self.pcls_file = pcls_file
    self.signatures_dir = signatures_dir

    if self.verbose:
      print("Loading reference dataset; this may take a while...")
    self.cmap = self.read_reference_dataset()
    if self.verbose:
      print("...done.")
    self.signatures = self.read_signatures()

    self.init_connectivity()

    if self.verbose:
      print("Loading gene symbol data.")
    # Handle gene symbols
    self.hsap_symbol_map = np.load(SYMBOL_MAP_FNAME)
    self.cmap_genes = self.process_cmap_genes()

    if self.verbose:
      print("Parsing perturbagen class (PCL) data.")
    with open(self.pcls_file, 'r') as fp:
      self.pcls = json.load(fp)

    if self.verbose:
      print("Data structures initialized - VenomSeq is now ready for use!")

  def __repr__(self):
    """Return a string that summarizes the data loaded into the VenomSeq
    object. This should look something like the output of `summary()` in R.
    """
    return """VenomSeq object of {0} venom signatures.
    """.format(
      len(self.signatures)
    )

  def init_connectivity(self):
    ConnectivityData = namedtuple('ConnectivityData', ['wcs','ncs','tau'])
    self.connectivity = ConnectivityData(wcs=None, ncs=None, tau=None)

  def load(self,
           wcs_file=None,
           ncs_file=None,
           tau_file=None):
    """Load precomputed VenomSeq data from local files.
    """
    if not (wcs_file or ncs_file or tau_file):
      raise Exception('Must supply at least one filename argument to load().')

    if wcs_file is not None:
      self.connectivity = self.connectivity._replace(wcs=np.load(wcs_file))
    if ncs_file is not None:
      self.connectivity = self.connectivity._replace(ncs=np.load(ncs_file))
    if tau_file is not None:
      self.connectivity = self.connectivity._replace(tau=np.load(tau_file))

  def subset_connectivity_data(self, type='tau', cell_line=None, pert_type='trt_cp',
                               top_n=None, top_n_metric='var'):
    """Retrieve a subset of a data matrix matching certain conditions.

    Parameters
    ----------
    type : str
      The connectivity data matrix to use. It can be 'wcs', 'ncs', or 'tau';
      other options will be added in future releases.
    cell_line : str, optional
      Cell line name to filter for. If omitted, all cell lines will be returned.
    pert_type : str, optional
      Perturbagen type to filter for. If omitted, all perturbagen types will be
      returned.
    top_n : int, optional
      Provide this argument if you would like only the top N features (connectivites)
      across all venoms with regards to a certain metric. This is especially useful
      for filtering out features with low variance, which can make visualizations
      and clusterings much more challenging to interpret. 
    top_n_metric : str, optional
      String argument corresponding to the metric used for determining the top_n
      features. Changing this parameter only has an effect if `top_n` is not `None`.

    Returns
    -------
    subset : numpy.ndarray
      Returns a 2-dimensional Numpy array containing the requested data.
    """
    if type == 'wcs':
      subset = self.connectivity.wcs
    elif type == 'ncs':
      subset = self.connectivity.ncs
    elif type == 'tau':
      subset = self.connectivity.tau

    cell_idxs = np.arange(subset.shape[0])
    pert_idxs = np.arange(subset.shape[0])

    if cell_line:
      cell_idxs = self._cell_type_idxs(cell_line)
    if pert_type:
      pert_idxs = self._pert_type_idxs(pert_type)

    filter_idxs = np.intersect1d(cell_idxs, pert_idxs)

    subset = subset[filter_idxs,:]

    if top_n: 
      metric_values = np.var(subset, axis=1)
      metric_ranks = stats.rankdata(metric_values)
      greatest_n = metric_ranks.argsort()[-(top_n):]
      subset = subset[greatest_n,:]

    return subset

  def pcl_signatures_with_annotations(self, type='tau', top_n=None, top_n_metric='var', cell_line='A375', pert_type='trt_cp'):
    if type == 'wcs':
      subset = self.connectivity.wcs
    elif type == 'ncs':
      subset = self.connectivity.ncs
    elif type == 'tau':
      subset = self.connectivity.tau

    # Get indices for pcl sig ids
    pcl_sig_ids = self._pcl_idxs()
    sig_idx_mask = self.cmap.cols.index.isin(pcl_sig_ids)
    sig_idx = np.where(np.array(sig_idx_mask))[0]

    # If we need to mask cell types and pert types:
    cell_idxs = np.arange(subset.shape[0])
    pert_idxs = np.arange(subset.shape[0])
    if cell_line:
      cell_idxs = self._cell_type_idxs(cell_line)
    if pert_type:
      pert_idxs = self._pert_type_idxs(pert_type)

    filter_idxs = np.intersect1d(cell_idxs, pert_idxs)
    filter_idxs = np.intersect1d(filter_idxs, sig_idx)

    subset = subset[filter_idxs,:]
    pcl_annotations = self.cmap.cols.iloc[filter_idxs,:]

    if top_n:
      metric_values = np.var(subset, axis=1)
      metric_ranks = stats.rankdata(metric_values)
      greatest_n = metric_ranks.argsort()[-(top_n):]
      subset = subset[greatest_n,:]
      pcl_annotations = pcl_annotations.iloc[greatest_n,:]

    return (subset, pcl_annotations)

  def pcls_from_annotations(self, annot):
    """Given a DataFrame of annotation data corresponding to connectivity map
    profiles (e.g., the second value returned by `pcl_signatures_with_annotations()`),
    this method returns PCL membership for each of the signatures in that list, in
    order. This is useful, for example, when you want to annotate rows in a clustermap
    with PCL membership to see if PCLs cluster together.

    Parameters
    ----------
    annot : pd.DataFrame
      The data frame corresponding to a block of connectivity scores. These should be
      sorted in the same order as the rows (0'th axis) of the score matrix. The index
      of this dataframe should be the signature IDs.
    """
    # To reduce time complexity, we first create a boolean vector that indicates
    # whether the signature is in ANY PCL. If it is, we can then loop through
    # all PCLs to determine membership. This effectively skips evaluation of
    # the majority of signatures.
    pcl_sig_ids = self._pcl_idxs()
    sig_idx_mask = self.cmap.cols.index.isin(pcl_sig_ids)
    sigs_in_pcls = set(self.cmap.cols.index[sig_idx_mask])

    pcl_sig_id_index = {}
    for p in self.pcls:
      unjoined = [x['sig_id'] for x in p['perts']]
      joined = list(itertools.chain.from_iterable(unjoined))
      pcl_sig_id_index[p['group_id']] = joined
    
    pcl_membership = []
    for sig_id in tqdm(annot.index):
      if sig_id in sigs_in_pcls:
        this_membership = []
        # loop through PCLs and test membership
        for pcl in self.pcls:
          if sig_id in pcl_sig_id_index[pcl['group_id']]:
            this_membership.append(pcl['group_id'])
      pcl_membership.append(this_membership)
      
    assert (len(pcl_membership) == annot.shape[0])
    
    return pcl_membership

  def _pcl_idxs(self):
    all_pcl_sig_ids = []
    
    for pcl in self.pcls:
      this_pcl_sig_ids = [x['sig_id'] for x in pcl['perts']]
      this_pcl_sig_ids = list(itertools.chain.from_iterable(this_pcl_sig_ids))
      all_pcl_sig_ids.append(this_pcl_sig_ids)
    all_pcl_sig_ids = list(itertools.chain.from_iterable(all_pcl_sig_ids))
    all_pcl_sig_ids = np.array(list(set(all_pcl_sig_ids))) # Remove duplicates
    
    return all_pcl_sig_ids

  def _cell_type_idxs(self, cell_type):
    mask = np.array(self.cmap.cols.cell_id == cell_type)
    return np.nonzero(mask)[0]

  def _pert_type_idxs(self, pert_type):
    mask = np.array(self.cmap.cols.pert_type == pert_type)
    return np.nonzero(mask)[0]

  def process_cmap_genes(self):
    """Parse an integer-valued list of NCBI gene IDs corresponding
    to the individual rows of the CMap data table.

    GCTX metadata tables don't always conform to a standard, so this
    code may necessarily become messy to account for edge-cases as
    they are encountered.
    """
    cmap_genes = self.cmap.rows.index
    if cmap_genes[0][-3:] == '_at':
      # We need to look in a different column for the gene IDs
      cmap_genes = self.cmap.rows['pr_gene_id'].astype(int).tolist()
    return cmap_genes

  def read_reference_dataset(self):
    CMap = namedtuple('CMap', ['data', 'cols', 'rows'])
    data_df, cm, rm = read_gctx(self.gctx_file, ignore_data_df=True)
    return CMap(data=data_df, cols=cm, rows=rm)

  def read_signatures(self):
    sig_files = glob("{0}/*.csv".format(self.signatures_dir))
    sigs_pd = []
    sigs = []

    for f in sig_files:
      v = f.split("/")[-1].split(".")[0]
      sig = pd.read_csv(f, sep=",")
      sig = sig[pd.notnull(sig['symbol'])]
      sigs_pd.append((v, sig))

    for s in sigs_pd:
      i, sig = s
      n_down = sig.loc[sig['log2FoldChange'] < 0].shape[0]
      n_up = sig.loc[sig['log2FoldChange'] > 0].shape[0]
      sigs.append({
        'venom': i,
        'n_up': n_up,
        'n_down': n_down,
        'up': np.array(sig.loc[sig['log2FoldChange'] > 0]),
        'up_list': list(np.array(sig.loc[sig['log2FoldChange'] > 0])[:,1]),
        'down': np.array(sig.loc[sig['log2FoldChange'] < 0]),
        'down_list': list(np.array(sig.loc[sig['log2FoldChange'] < 0])[:,1]),
      })

    return sigs

  def compute_connectivities(self):
    pass

  def normalize_connectivities(self):
    pass

  def compute_taus(self):
    pass

  def compute_pcl_enrichments(self):
    pass


import numpy as np

from cmapPy.pandasGEXpress.parse import parse

def read_gctx(fname, col_meta=True, row_meta=True, ignore_data_df=False):
  print("  Parsing GCTX file.")

  if ignore_data_df:
    print("  Loading row metadata.")
    rm = parse(fname, row_meta_only=True)
    fix_mangled_byte_literals(rm)
    print("  Loading column metadata.")
    cm = parse(fname, col_meta_only=True)
    fix_mangled_byte_literals(cm)
    cm['sig_num'] = list(range(cm.shape[0]))

    return (None, cm, rm)

  else:
    # Load everything
    print("  IGNORING EXPRESSION SIGNATURES; ONLY LOADING METADATA.")
    print("  If you did not intend to do this, re-run with different arguments.")
    tmp = parse(fname)
    data_df = tmp.data_df

    print("  Fixing mangled byte literals.")
    fix_mangled_byte_literals(data_df)

    if (col_meta):
      print("  Loading column metadata")
      cm = tmp.col_metadata_df
      fix_mangled_byte_literals(cm)
      cm['sig_num'] = list(range(cm.shape[0]))
    if (row_meta):
      print("Loading row metadata")
      rm = tmp.row_metadata_df
      fix_mangled_byte_literals(rm)

    return (data_df, cm, rm)


def fix_mangled_byte_literals(data_df):
  columns = data_df.columns
  idx = data_df.index

  if columns[0][:2] == "b'":
    data_df.columns = [c[2:-1] for c in columns]
  if idx[0][:2] == "b'":
    data_df.index = [i[2:-1] for i in idx]

  # check if byte literals exist in the cells (e.g., if the dataframe contains metadata)
  if data_df.shape != data_df.select_dtypes(np.number).shape:
    # At least one column is nonnumeric, and needs to be checked
    for col in data_df.columns:
      colvalues = np.array(data_df.loc[:,col])
      if not np.issubdtype(colvalues.dtype, np.number):
        if colvalues[0][:2] == "b'":
          data_df.loc[:,col] = [cv[2:-1] for cv in colvalues]



#!/usr/bin/env python

import setuptools

MAJOR       = 0
MINOR       = 1
PRERELEASE  = 'alpha'
ISRELEASED  = False
if PRERELEASE:
  VERSION = '%d.%d-%s' % (MAJOR, MINOR, PRERELEASE)
else:
  VERSION = '%d.%d' % (MAJOR, MINOR)

with open("README.md", "r") as fp:
  long_description = fp.read()

setuptools.setup(
  name='VenomSeq',
  version=VERSION,
  description='Python code for performing the data analysis in the VenomSeq workflow',
  long_description=long_description,
  author='Joseph D. Romano',
  author_email='jdromano2@gmail.com',
  url='https://github.com/jdromano2/venomseq',
  packages=['venomseq'],
  package_dir={'venomseq': 'venomseq'},
  classifiers=[
    "Development Status :: 3 - Alpha",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Programming Language :: Python :: 3.6",
    "Topic :: Scientific/Engineering :: Bio-Informatics"
  ]
)


import numpy as np
import scipy as sp
from tqdm import tqdm
import time
from collections import defaultdict

import ipdb

# Default metaparameters
BLOCK_NCOLS = 1000

class Algorithm(object):
  """Base class stub for all algorithms for analyzing VenomSeq data.
  """
  def __init__(self, venomseq):
    self.venomseq = venomseq


class Connectivity(Algorithm):
  """Connectivity Analysis algorithm for VenomSeq data.
  """
  def __init__(self,
               venomseq,
               block_ncols=BLOCK_NCOLS,
               verbose=True):
    super(Connectivity, self).__init__(venomseq=venomseq)
    self.verbose = verbose

    self.block_ncols = block_ncols

  def run(self):
    if self.verbose:
      print("Beginning connectivity analysis algorithm.")

    self.venoms = []
    #self.dim_cmap = self.venomseq.cmap.data.shape
    self.dim_cmap = (self.venomseq.cmap.rows.shape[0], self.venomseq.cmap.cols.shape[0])

    self.chunk_idxs = self.get_col_blocks(self.dim_cmap, (self.dim_cmap[0], self.block_ncols))

    # Perform main steps of the algorithm
    self.compute_wcs()
    self.compute_genewise_statistics()
    self.normalize_connectivities()
    self.tau_quantize()

    # Copy data to VenomSeq instance
    self.venomseq.connectivity = self.venomseq.connectivity._replace(wcs=self.wcs)
    self.venomseq.connectivity = self.venomseq.connectivity._replace(ncs=self.ncs)
    self.venomseq.connectivity = self.venomseq.connectivity._replace(tau=self.tau)

  def compute_wcs(self):
    if self.verbose:
      print("Computing weighted connectivity scores (WCSs); this may take a while.")

    wcs_start = time.time()
    wcs = np.zeros((self.dim_cmap[1], len(self.venomseq.signatures)))

    # For each venom...
    for v_idx, v_sig in tqdm(enumerate(self.venomseq.signatures), disable=(not self.verbose)):
      self.venoms.append(v_sig['venom'])

      # One venom's worth of WCSs
      scores = np.zeros(self.dim_cmap[1])

      up_ids = self.symbols_2_ids(v_sig['up_list'])
      down_ids = self.symbols_2_ids(v_sig['down_list'])

      # If a gene isn't in the CMap data, remove it
      up_ids_filter = self.filter_cmap_genes(up_ids)
      down_ids_filter = self.filter_cmap_genes(down_ids)

      # For each block of reference signatures...
      for chunk_col_idx in tqdm(self.chunk_idxs, disable=(not self.verbose)):
        c_start, c_end = chunk_col_idx
        chunk = np.array(self.venomseq.cmap.data.iloc[:,c_start:(c_end+1)])

        # For each signatures within the chunk...
        for i in range(chunk.shape[-1]):
          ranks = self.ranked_ids(chunk[:,i])
          ss = self.score_signature(ranks, up_ids_filter, down_ids_filter)
          scores[chunk_col_idx[0]+i] = ss

      wcs[:,v_idx] = scores

    wcs_end = time.time()
    wcs_delta = wcs_end - wcs_start

    if self.verbose:
      print("Computed {0} WCSs in {1} second(s).".format(wcs.shape[0]*wcs.shape[1], wcs_delta))

    # TODO: Make the transpose unnecessary to alleviate future confusion
    self.wcs = wcs.T

  def compute_genewise_statistics(self):
    if self.verbose:
      print("Computing means for cell type and perturbagen type combinations")

    mu = defaultdict(dict)
    all_cell_types = set(self.venomseq.cmap.cols['cell_id'])
    all_pert_types = set(self.venomseq.cmap.cols['pert_type'])
    for c in tqdm(all_cell_types, disable=(not self.verbose)):
      for t in all_pert_types:
        sub_df = self.venomseq.cmap.cols[ \
          (self.venomseq.cmap.cols['cell_id']==c) & \
          (self.venomseq.cmap.cols['pert_type']==t)]

        if sub_df.shape[0] == 0:
          mu[c][t] = (None, None)
        else:
          #ipdb.set_trace()
          sub_idxs = np.array(sub_df['sig_num'])
          sub = self.wcs[:,sub_idxs]
          mu_pos = np.mean(sub[sub > 0.])
          mu_neg = np.mean(sub[sub < 0.])
          mu[c][t] = (mu_pos, mu_neg)
    self.mu = mu

  def normalize_connectivities(self):
    if self.verbose:
      print("Normalizing connectivities by cell line and perturbagen type.")

    ncs_ct = np.zeros(self.wcs.shape)

    # x: CMap signature index
    for x in tqdm(range(self.wcs.shape[-1]), total=self.wcs.shape[-1]):
      cell_type = self.venomseq.cmap.cols.iloc[x]['cell_id']
      pert_type = self.venomseq.cmap.cols.iloc[x]['pert_type']
      mus = self.mu[cell_type][pert_type]

      #normalized_col = [] <-- Did I need this...?
      # y: venom number
      for y in range(self.wcs.shape[0]):
        w = self.wcs[y,x]
        if w > 0.:
          ncs_ct[y,x] = w / mus[0] # positive
        elif w < 0.:
          ncs_ct[y,x] = -(w / mus[1]) # negative

    self.ncs = ncs_ct

  def tau_quantize(self):

    if self.verbose:
      print("Computing Tau scores from normalized connectivity scores.")

    x, y = np.unique(self.venomseq.cmap.cols['cell_id'], return_counts=True)
    kept_cell_lines = list(x[y >= 10])

    all_taus = {}

    for cl in tqdm(kept_cell_lines, disable=(not self.verbose)):
      cur_cell_idxs = np.array((self.venomseq.cmap.cols.loc[self.venomseq.cmap.cols['cell_id'] == cl].sig_num))
      cur_cell_ncs = self.ncs[:,cur_cell_idxs]
      cur_cell_abs = np.abs(cur_cell_ncs).flatten() # Scipy's quantile score freaks out if you don't flatten
      cur_cell_taus = np.zeros_like(cur_cell_ncs)

      for i in tqdm(range(0, cur_cell_ncs.shape[0]), leave=False, disable=(not self.verbose)):
        for j in tqdm(range(0, cur_cell_ncs.shape[1]), leave=False, disable=(not self.verbose)):
          sgn = np.sign(cur_cell_ncs[i,j])
          tau = sgn * sp.stats.percentileofscore(cur_cell_abs, np.abs(cur_cell_ncs[i,j]))
          cur_cell_taus[i,j] = tau

      all_taus[cl] = {
        'idxs': cur_cell_idxs,
        'taus': cur_cell_taus
      }

    t = np.zeros_like(self.ncs)
    for _, data in all_taus.items():
      t[:,data['idxs']] = data['taus']

    self.tau = t


  def ranked_ids(self, drug_arr, desc=True):
    """Obtain a list of ranked genes for a drug in Connectivity Map. The values
    of the array returned by this function are the Entrez gene ids in the
    ranked order.

    Keyword arguments:
    drug_arr -- a numpy.array of normalized gene values (e.g., Level 5 data)
    desc -- Boolean; return with descending ranks (if false, return ascending)
    """
    assert (len(drug_arr.shape) == 1)

    rankidxs = np.argsort(drug_arr)
    ranked = np.empty_like(rankidxs, dtype=int)
    ranked[rankidxs] = self.venomseq.cmap_genes

    if desc:
      ranked = np.flip(ranked, axis=0)
    return ranked

  def enrichment_score(self, drug_ranked, enrichment_set, up=True):
    """Compute an enrichment score, as done by Sirota et al and Lamb et al."""
    r = len(drug_ranked)     # 'probe set'; aka `n`
    s = len(enrichment_set)  # 'tags'; aka `t`
    # V: the indexes of the reference set that are in the enrichment set
    V = np.empty_like(enrichment_set)

    # for each gene in the enrichment set...
    for j in range(s):
      # find the index of that gene in the reference set
      matched = np.argwhere(drug_ranked == enrichment_set[j])
      assert (len(matched) == 1)
      # insert that value into V
      V[j] = matched[0]

    V.sort

    a = 0.
    b = 0.
    # in Lamb et al's definition of a and b, be careful with redefined variables
    for j2 in range(s):
      a_tmp = (j2/s) - (V[j2]/r)
      b_tmp = (V[j2]/r) - ((j2-1)/s)
      a = max(a, a_tmp)
      b = max(b, b_tmp)

    if a > b:
      return a
    else:
      return -b

  def symbols_2_ids(self, symbols):
    vsm = self.venomseq.hsap_symbol_map
    return [int(np.squeeze(vsm[vsm[:,0] == x,:])[1]) for x in symbols]

  def filter_cmap_genes(self, comp_ids):
    mask = np.in1d(comp_ids, self.venomseq.cmap_genes)
    ids_filter = np.extract(mask, comp_ids)
    return ids_filter

  def score_signature(self, drug_ranked, up, down):
    """Given a CMap drug and a comparison signature, compute a connectivity score
    as described in Sirota et al and Lamb et al.

    Keyword arguments:
    drug_ranked -- An int-valued numpy array of rank-sorted Entrez gene ids
    comp_sig -- A dictionary containing a signature to compare, formatted as
                shown above (in `sigs`)
    """
    assert (len(drug_ranked.shape) == 1)

    up_c = np.copy(up)
    down_c = np.copy(down)

    es_up = self.enrichment_score(drug_ranked, up_c)
    es_down = self.enrichment_score(drug_ranked, down_c)

    if (((es_up== es_down) & (es_up==0)) | (es_up*es_down>0)):
      return 0
    else:
      return (es_up - es_down)

  def get_col_blocks(self, total_shape, block_shape):
    """Return indices corresponding to blocks that iterate
    over columns."""
    out_shape_tups = []

    block_ncols = block_shape[-1]
    #total_shape = cmap_data.shape
    block_shape = (total_shape[0], block_ncols)
    total_shape = np.array(total_shape)
    block_shape = np.array(block_shape)

    remainder = 0

    if (total_shape % block_shape).sum() != 0:
      remainder = (total_shape % block_shape)[-1]
      n_cols_trim = total_shape[1] - remainder
    else:
      n_cols_trim = total_shape[1]

    n_blocks = int(n_cols_trim / block_ncols)
    for i in range(n_blocks):
      out_shape_tups.append( (i*block_ncols, ((i+1)*block_ncols)-1) )

    if remainder > 0:
      out_shape_tups.append( (total_shape[-1]-remainder, total_shape[-1]-1) )

    return out_shape_tups
